# -*- coding: utf-8 -*-
"""
Created on Fri Jun 26 17:12:52 2020

@author: kourf

"""

import random
import tree
from graphviz import Digraph


class Initialisation:
    def __init__(self):
        cards2pick = list(range(1, 11))*2 + [25, 50, 75, 100]
        self.__cards = random.sample(cards2pick, 6)
        self.__goal = random.randint(100, 999)

    def get_cards(self):
        return self.__cards

    def get_goal(self):
        return self.__goal

class Number(tree.Node):
    def __init__(self, data):
        "data est un nombre entier"
        super().__init__(data, None, None)

    def __repr__(self):
        return "{}".format(self._Node__data)

    def get_value(self):
        return self._Node__data


class Operator(tree.Node):
    def __init__(self, op, symbol, left, right):
        """
        op est une fonction à deux variables x et y renvoyant
        le résultat d'une opération arithmétique +, -, * ou /.
        symbol est le symbole correspondant parmi '+', '-', '*', '/'
        """
        super().__init__((op, symbol), left, right)
        self.__op = self._Node__data[0]
        self.__symbol = self._Node__data[1]
        # Correction de la Q4a
        self.__value = self.eval()
        # Fin de correction de la Q4a (suite plus bas)

    def get_op(self):
        return self.__op

    def get_symbol(self):
        return self.__symbol
      
    # Correction de la Q4a (suite)
    def eval(self):
        return self.get_op()(self.left().get_value(), self.right().get_value())
    
    def get_value(self):     
        return self.__value
    # Fin correction de la Q4a
    
    
    # Correction de la Q4c    
    def __repr__(self):
       if self.__symbol in ['+', '-']: chaine = "({} {} {})"
       else: chaine = "{} {} {}"
       return chaine.format(self.left().__repr__(), self.get_symbol(), self.right().__repr__())
    # Fin de la correction de la Q4c

"""
La méthode eval est appelée une seule fois lors de la construction de 
l'opérateur. Le résultat est stocké dans la propriété __value, ce qui 
évite de recalculer l'évaluation chaque fois qu'on en a besoin.
"""

class SolverCompteEstBon:
    def __init__(self, cards, goal):
        self.__cards = [Number(card) for card in cards]
        self.__goal = goal
        self.__operators = [
                ((lambda x, y: x + y), '+'),
                ((lambda x, y: x * y), '*'),
                ((lambda x, y: x - y), '-'),
                ((lambda x, y: x//y), '/')
                ]

    def make_expressionTree_list(self):

        def me_rec(cards, goal):
            if len(cards) > 1:
                n = len(cards)
                for i in range(n-1):
                    for j in range(i+1,n):
                        for op, symbol in self.__operators:
                            left = cards[i]
                            right = cards[j]
                            # Correction de la Q4b
                            if symbol=='/' or symbol=='-':
                               if left.get_value() < right.get_value():
                                   left, right = right, left
                               if symbol=='/' and (right.get_value()==0 or
                                       left.get_value()%right.get_value()!=0):
                                   continue
                            # Fin de correction de la Q4b
                            newcard = Operator(op, symbol, left, right)
                            if newcard.get_value() == goal:
                                expressionTree_list.append(newcard)
                            else:
                                newcards = [newcard]+cards[:i]+cards[i+1:j]+cards[j+1:]
                                me_rec(newcards, goal)

        expressionTree_list = []
        me_rec(self.__cards, self.__goal)
        return expressionTree_list


# ******************************************************************************
#
#               DO NOT CHANGE ABOVE EXCEPT FOR Q4
#
# ******************************************************************************


def showtree(tree):
    def rec_construction(tree, dot):
        mother_id = str(node_id[0])
        if isinstance(tree, Operator):
            symbol = tree.get_symbol()
            for d in (tree.left(), tree.right()):
                node_id[0] += 1
                if isinstance(d, Operator):
                    dot.node(str(node_id[0]), label=d.get_symbol())
                else:
                    dot.node(str(node_id[0]), label=str(d.get_value()))
                dot.edge(mother_id, str(node_id[0]))
                rec_construction(d, dot)
    dot = Digraph('G', filename='expression' )
    node_id = [0]
    dot.node(str(node_id[0]), label=tree.get_symbol())
    rec_construction(tree, dot)
    dot.view()
    return dot
            
        
# ******************************************************************************
#
#               WRITE BELOW
#
# ******************************************************************************



# ******************************************************************************
#
#               DO NOT DELETE BELOW
#
# ******************************************************************************


if __name__=='__main__':
    game = Initialisation()
    cards = game.get_cards()
    goal = game.get_goal()
    for c in cards: print(c, end='  ')
    print('\n',goal,sep='')
    expressions = SolverCompteEstBon(cards, goal).make_expressionTree_list()










