# -*- coding: utf-8 -*-
"""
Created on Fri Jun 26 17:12:52 2020

@author: kourf

"""

import random
import tree
from graphviz import Digraph
import tkinter as tk
import time


class Initialisation:
    def __init__(self):
        cards2pick = list(range(1, 11))*2 + [25, 50, 75, 100]
        self.__cards = random.sample(cards2pick, 6)
        self.__goal = random.randint(100, 999)

    def get_cards(self):
        return self.__cards

    def get_goal(self):
        return self.__goal

class Number(tree.Node):
    def __init__(self, data):
        "data est un nombre entier"
        super().__init__(data, None, None)

    def __repr__(self):
        return "{}".format(self._Node__data)

    def get_value(self):
        return self._Node__data


class Operator(tree.Node):
    def __init__(self, op, symbol, left, right):
        """
        op est une fonction à deux variables x et y renvoyant
        le résultat d'une opération arithmétique +, -, * ou /.
        symbol est le symbole correspondant parmi '+', '-', '*', '/'
        """
        super().__init__((op, symbol), left, right)
        self.__op = self._Node__data[0]
        self.__symbol = self._Node__data[1]
        # Correction de la Q4a
        self.__value = self.eval()
        # Fin de correction de la Q4a (suite plus bas)

    def get_op(self):
        return self.__op

    def get_symbol(self):
        return self.__symbol
      
    # Correction de la Q4a (suite)
    def eval(self):
        return self.get_op()(self.left().get_value(), self.right().get_value())
    
    def get_value(self):     
        return self.__value
    # Fin correction de la Q4a
    
    
    # Correction de la Q4c    
    def __repr__(self):
       if self.__symbol in ['+', '-']: chaine = "({} {} {})"
       else: chaine = "{} {} {}"
       return chaine.format(self.left().__repr__(), self.get_symbol(), self.right().__repr__())
    # Fin de la correction de la Q4c

"""
La méthode eval est appelée une seule fois lors de la construction de 
l'opérateur. Le résultat est stocké dans la propriété __value, ce qui 
évite de recalculer l'évaluation chaque fois qu'on en a besoin.
"""

class SolverCompteEstBon:
    def __init__(self, cards, goal):
        self.__cards = [Number(card) for card in cards]
        self.__goal = goal
        self.__operators = [
                ((lambda x, y: x + y), '+'),
                ((lambda x, y: x * y), '*'),
                ((lambda x, y: x - y), '-'),
                ((lambda x, y: x//y), '/')
                ]

    def liste_calculs(self, expression):
        """
        Renvoie la liste de calculs d'une expression
        """
        def liste_calculs_(expression):
            if isinstance(expression, Operator):
                liste_calculs_(expression.left())
                liste_calculs_(expression.right())
                res.append(str(expression.left().get_value()) + 
                                expression.get_symbol() + 
                                str(expression.right().get_value()) + 
                                '=' + str(expression.get_value()))

        res = []
        liste_calculs_(expression)
        return res
        


    def make_expressionTree_list(self):
        def me_rec(cards, goal):
            if len(cards) > 1:
                n = len(cards)
                for i in range(n-1):
                    for j in range(i+1,n):
                        for op, symbol in self.__operators:
                            left = cards[i]
                            right = cards[j]
                            # Correction de la Q4b
                            if symbol=='/' or symbol=='-':
                               if left.get_value() < right.get_value():
                                   left, right = right, left
                               if symbol=='/' and (right.get_value()==0 or
                                       left.get_value()%right.get_value()!=0):
                                   continue
                            # Fin de correction de la Q4b
                            newcard = Operator(op, symbol, left, right)
                            if newcard.get_value() == goal:
                                expressionTree_list.append(newcard)
                            else:
                                newcards = [newcard]+cards[:i]+cards[i+1:j]+cards[j+1:]
                                me_rec(newcards, goal)

        expressionTree_list = []
        me_rec(self.__cards, self.__goal)
        return expressionTree_list

    def elimine_doublons(self, expressions):
        """
        Élimine les vrais doublons, mais conserve les doublons du type 
            2+3=5, 4+6=10, 5*10=50 
        et  4+6=10, 2+3=5, 10*5=50 
        Il ne les reconnait pas comme doublons car '5*10' != '10*5'
        Ce problème est évité avec les cartes initiales, le solveur tenant 
        compte de la commutativité, mais pas en deuxième étape.
        On ne tient pas compte de l'associativité : 
        2*5 = 10, 10*3 = 30 est différent de 2*3 = 6, 6*5 = 30.
        """
        expressions_calculs = []
        for e in expressions:
            lst_calc = sorted(self.liste_calculs(e))
            expressions_calculs.append((lst_calc, e))
        expressions_calculs.sort(key=lambda x: x[0])
        lst_calc = expressions_calculs[0][0]
        expressions_sans_doublons = [expressions_calculs[0][1]]
        for (c, e) in expressions_calculs:
            if c != lst_calc:
                lst_calc = c
                expressions_sans_doublons.append(e)
        return expressions_sans_doublons


    def solve(self):
        expressions = self.make_expressionTree_list()
        if expressions:
            expressions = self.elimine_doublons(expressions)
        return expressions
            
            



# ******************************************************************************
#
#               DO NOT CHANGE ABOVE EXCEPT FOR Q4
#
# ******************************************************************************




def showtree(tree):
    def rec_construction(tree, dot):
        mother_id = str(node_id[0])
        if isinstance(tree, Operator):
            symbol = tree.get_symbol()
            for d in (tree.left(), tree.right()):
                node_id[0] += 1
                if isinstance(d, Operator):
                    dot.node(str(node_id[0]), label=d.get_symbol())
                else:
                    dot.node(str(node_id[0]), label=str(d.get_value()))
                dot.edge(mother_id, str(node_id[0]))
                rec_construction(d, dot)
    dot = Digraph('G', filename='expression' )
    node_id = [0]
    dot.node(str(node_id[0]), label=tree.get_symbol())
    rec_construction(tree, dot)
    dot.view()
    return dot
            
        
# ******************************************************************************
#
#               WRITE BELOW
#
# ******************************************************************************


class Entry_num(tk.Entry):
    def __init__(self, master, color='black'):
        self.__txt = tk.StringVar()
        tk.Entry.__init__(self, master, textvariable=self.__txt, width=4, font = "Helvetica 18", justify="center", fg=color)
        self.__txt.set(str(0))
        self.__value = 0
        self.__txt.trace_add('write', self.modify)

    def get_value(self):
        return self.__value
    
    def set_value(self, value):
        self.__value = value
        self.__txt.set(str(value))


    def modify(self, *args):
        """
        Attention les variables goal et cards ne sont pas actualisées
        lorsqu'on modifie à la main les valeurs affichées. L'actualisation
        se fait au moment de lancer le solveur.
        """
        try:
            self.__value = int(self.__txt.get())
        except:
            self.__txt.set(self.__value)
                


class Timer(tk.Label):
    def __init__(self, master):
        tk.Label.__init__(self, master, text="1:00", font ="Helvetica 16 bold")
        self.__start = time.time()
        self.update_clock()
        
    def update_clock(self):
        now = time.time()
        t = now - self.__start
        if 1 <= t:
            if t <= 51:
                counter = '0:'+str(60 - int(t))
            else:
                counter = '0:0'+str(60 - int(t))
            self.configure(text=counter)
        if t > 60:
            self.configure(text="0:00")
        else:
            self.master.after(999, self.update_clock)
            
    def reset(self):
        self.__start = time.time()
        self.configure(text="1:00")
        self.update_clock()


class IHMGfx(tk.Tk):
    def __init__(self):
        tk.Tk.__init__(self)    
        self.title("Le compte est bon")
        self.__disp_cards = [0]*6
        for i in range(6):
            self.__disp_cards[i] = Entry_num(self)
            self.__disp_cards[i].grid(row=0, column=i, padx=10, pady=10)
        self.__disp_goal = Entry_num(self, color='red')
        self.__disp_goal.grid(row=1, column=4, padx=10, pady=10)
        self.__cards, self.__goal = self.tirage_valeurs(self.__disp_cards, self.__disp_goal)
        self.__nbrSolutions = 0
        tk.Button(self, text ="Nouveau tirage", command =self.reroll).grid(row=2, column=0, columnspan=3)
        tk.Button(self, text ="Solution", command =self.solve).grid(row=2, column=3, columnspan=3)
        tk.Label(self, text="Horloge", font = "Helvetica 16 bold").grid(row=3, column=0, columnspan=2, padx=10, pady=10, sticky='w')
        self.__timer = Timer(self)
        self.__timer.grid(row=3, column=2, columnspan=2, padx=10, pady=10, sticky='w')
        tk.Label(self, text="Solution", font = "Helvetica 16 bold").grid(row=4, column=0, columnspan=5, padx=10, pady=10, sticky='w')
        self.__solchaine = tk.StringVar()
        tk.Label(self, textvariable=self.__solchaine, font = "Helvetica 18", width=30, anchor='w', relief='ridge').grid(row=5, column=0, columnspan=6, padx=10, pady=5, sticky='w')
        self.__solcalculs = tk.StringVar()
        tk.Label(self, textvariable=self.__solcalculs, font = "Helvetica 18", width=30, anchor='w', relief='ridge').grid(row=6, column=0, columnspan=6, padx=10, pady=5, sticky='w')
        self.__solcalculs.set("\n \n \n \n \n \n \n")
        self.__solchaine.set("")
        self.__scale = tk.Scale(self, length=400, orient='horizontal', label ='Choix de la solution',
                  from_=0, to=0, command=self.disp_sol)
        self.__scale.grid(row=7, column=0, columnspan=6)
        self.__expresions = []
        
  
    def reroll(self):
        self.__cards, self.__goal = self.tirage_valeurs(self.__disp_cards, self.__disp_goal)
        self.__timer.reset()
        self.__scale.configure(to=0)
        self.__solcalculs.set("\n \n \n \n \n \n \n")
        self.__solchaine.set("")
        
    def tirage_valeurs(self, disp_cards, disp_goal):
        game = Initialisation()
        cards = game.get_cards()
        goal = game.get_goal()
        for i in range(6):
            disp_cards[i].set_value(cards[i])
        disp_goal.set_value(goal)
        return cards, goal
        

    def set_values(self):
        """
        Actualise les valeurs des variables cards et goal en utilisant
        les valeurs affichées
        """
        for i in range(6):
            self.__cards[i] = self.__disp_cards[i].get_value()
        self.__goal = self.__disp_goal.get_value()       

    def solve(self):
        self.set_values()
        self.__expressions = SolverCompteEstBon(self.__cards, self.__goal).solve()
        n = len(self.__expressions)
        if n==0:
                    self.__solcalculs.set("\n \n \n \n \n \n \n")
                    self.__solchaine.set("Pas de solution")
                    self.__scale.configure(to=0)
        else:
            self.disp_sol(0)
            self.__scale.configure(to=n-1)
        
    def disp_sol(self, num):
        if len(self.__expressions) > 0:
            num = int(num)
            self.__solchaine.set(self.__expressions[num])
            liste_calculs = SolverCompteEstBon(self.__cards, self.__goal).liste_calculs(self.__expressions[num])
            disp_calculs = ''
            for c in liste_calculs:
                disp_calculs += c + '\n'
            self.__solcalculs.set(disp_calculs)


# ******************************************************************************
#
#               DO NOT DELETE BELOW
#
# ******************************************************************************


if __name__=='__main__':
    IHMGfx().mainloop()
    

